/*--License:
	Kyra Sprite Engine
	Copyright Lee Thomason (Grinning Lizard Software) 2001-2002
	www.grinninglizard.com/kyra
	www.sourceforge.net/projects/kyra

	Kyra is provided under 2 licenses:

	- The GPL, with no additional restrictions.
	- The LGPL, provided you display the Kyra splash screen, described below.


--- GPL License --
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	The full text of the license can be found in license.txt


--- LGPL License --
  **Provided you kindly display the Kyra splash screen (details below), 
	you	may use the LGPL license:**

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	The full text of the license can be found in lgpl.txt


--- Kyra Splash Screen.

	It would be appreciate if you display the Kyra splash screen when using
	either license, however it is only required for the LGPL. All the
	resources for the splash are compiled into the library, and it can be
	accessed through the following API:

		KrEngine::StartSplash
		KrEngine::UpdateSplash
		KrEngine::EndSplash

	Full documentation is provided with the KrEngine class. The splash screen
	should be displayed for 2 seconds.

	Thank you.
*/

#ifndef GlFixed_MATH_INCLUDED
#define GlFixed_MATH_INCLUDED

#include <limits.h>
#include "gltypes.h"
#include "gldebug.h"

#define GlFixed_0 		( 0 )		// Help the compliler - use 1<<16 instead of a const.
#define GlFixed_1 		( 1<<16 )	// Help the compliler - use 1<<16 instead of a const.

inline S32		GlIntToFixed( int x )			{ return x << 16; }
inline S32		GlDoubleToFixed( double x )		{ return S32( x * 65536.0 + 0.5); }
inline double	GlFixedToDouble( S32 f )		{ return double( f ) / 65536.0; }
inline float	GlFixedToFloat( S32 f )			{ return float( f ) / 65536.0f; }
inline int      GlFixedToInt( S32 f )			{ return f >> 16; }
inline int		GlFixedToIntRound( S32 f )		{ return ( f + 0x0800 ) >> 16; }
inline int		GlFixedToIntRoundUp( S32 f )	{ return ( f + 0xffff ) >> 16; }


inline S32 GlFixedMult( S32 _x, S32 _y )
{
	#ifdef VALIDATE_DEBUG
		S64 x = _x;
		S64 y = _y;
		S64 ret = ( ( x * y ) >> 16 );
		GLASSERT( ret <= INT_MAX && ret >= INT_MIN );
		return S32( ret );
	#else
		return (S32) ( ( S64( _x ) * S64( _y ) ) >> 16 );
	#endif
}


inline S32 GlFixedDiv( S32 _x, S32 _y )
{
	#ifdef VALIDATE_DEBUG
		S64 x = ( S64( _x ) << 16 );
		S64 y = _y;

		S64 ret =  x / y;
		GLASSERT( ret <= INT_MAX && ret >= INT_MIN );
		return S32( ret );
	#else
		return S32( ( S64( _x ) << 16 ) / S64( _y ) );
	#endif
}


/**	A fixed math class. Uses 16.16 fixed point numbers.
*/
class GlFixed
{
  public:
	GlFixed()									{}

	GlFixed( const GlFixed &x )					{ v = x.v; }
	GlFixed( int x )							{ v = GlIntToFixed( x ); }
	GlFixed( const double x )					{ v = GlDoubleToFixed( x ); }

	#ifdef DEBUG
	void DebugDump()		{ GLOUTPUT( "fixed=%.2f\n", GlFixedToDouble( v ) ); }
	#endif


	// This could be done with overloaded operators, but I dislike overloaded
	// operators for casting because I think they are confusing when they
	// are used. -lee

	int    ToInt()	const					{ return GlFixedToInt( v ); }
	int    ToIntRoundUp()	const			{ return GlFixedToIntRoundUp( v ); }
	int    ToIntRound() const				{ return GlFixedToIntRound( v ); }
	double ToDouble() const					{ return GlFixedToDouble( v ); }
	float  ToFloat() const					{ return GlFixedToFloat( v ); }

	GlFixed Abs()							{ GlFixed t; ( v < 0 ) ? t.v = -v : t.v = v; return t; }

	GlFixed& operator = (const GlFixed &x)		{ v = x.v;          		return *this; }
	GlFixed& operator = (const int x)			{ v = GlIntToFixed( x );    return *this; }
	GlFixed& operator = (const double x)		{ v = GlDoubleToFixed( x ); return *this; }

	GlFixed& operator +=  (const GlFixed x)		{ v += x.v;          return *this; }
	GlFixed& operator +=  (const int x)			{ v += GlIntToFixed(x);    return *this; }
	GlFixed& operator +=  (const double x)		{ v += GlDoubleToFixed(x);    return *this; }

	GlFixed& operator -=  (const GlFixed x)		{ v -= x.v;          return *this; }
	GlFixed& operator -=  (const int x)			{ v -= GlIntToFixed(x);    return *this; }
	GlFixed& operator -=  (const double x)		{ v -= GlDoubleToFixed(x);    return *this; }

	GlFixed& operator *=  (const GlFixed x)		{ v = GlFixedMult(v, x.v);       return *this; }
	GlFixed& operator *=  (const int x)			{ v *= x;                     return *this; }
	GlFixed& operator *=  (const double x)		{ v = GlDoubleToFixed(GlFixedToDouble(v) * x);  return *this; }

	GlFixed& operator /=  (const GlFixed x)		{ v = GlFixedDiv(v, x.v);						return *this; }
	GlFixed& operator /=  (const int x)			{ v /= x;										return *this; }
	GlFixed& operator /=  (const double x)		{ v = GlFixedDiv( v, GlDoubleToFixed(x) );			return *this; }

	GlFixed& operator <<= (const int x)			{ v <<= x;           return *this; }
	GlFixed& operator >>= (const int x)			{ v >>= x;           return *this; }

	GlFixed& operator ++ ()						{ v += GlFixed_1;    return *this; }
	GlFixed& operator -- ()						{ v -= GlFixed_1;    return *this; }

	GlFixed operator ++ (int)						{ GlFixed t;  t.v = v;   v += GlFixed_1;  return t; }
	GlFixed operator -- (int)						{ GlFixed t;  t.v = v;   v -= GlFixed_1;  return t; }

	// Negation.
	GlFixed operator - () const					{ GlFixed t;  t.v = -v;  return t; }

	inline friend GlFixed operator +  (const GlFixed x, const GlFixed y)    { GlFixed t;  t.v = x.v + y.v; return t; }
	inline friend GlFixed operator +  (const int x,	   const GlFixed y)		{ GlFixed t;  t.v = GlIntToFixed(x) + y.v; return t; }
	inline friend GlFixed operator +  (const GlFixed x, const int y      )  { GlFixed t;  t.v = x.v + GlIntToFixed(y); return t; }

	inline friend GlFixed operator -  (const GlFixed x, const GlFixed y)    { GlFixed t;  t.v = x.v - y.v;        return t; }
	inline friend GlFixed operator -  (const int       x, const GlFixed y)    { GlFixed t;  t.v = GlIntToFixed(x) - y.v;        return t; }
	inline friend GlFixed operator -  (const GlFixed x, const int       y)    { GlFixed t;  t.v = x.v - GlIntToFixed(y);        return t; }
	inline friend GlFixed operator -  (const double x,    const GlFixed y)    { GlFixed t;  t.v = GlDoubleToFixed(x) - y.v;     return t; }
	inline friend GlFixed operator -  (const GlFixed x, const double    y)    { GlFixed t;  t.v = x.v - GlDoubleToFixed(y);     return t; }

	inline friend GlFixed operator *  (const GlFixed x, const GlFixed y)    { GlFixed t;  t.v = GlFixedMult(x.v, y.v);        return t; }
	inline friend GlFixed operator *  (const int x,       const GlFixed y)    { GlFixed t;  t.v = GlFixedMult(GlIntToFixed(x), y.v);        return t; }
	inline friend GlFixed operator *  (const GlFixed x, const int y)			 { GlFixed t;  t.v = GlFixedMult(x.v, GlIntToFixed(y));		return t; }
	inline friend GlFixed operator *  (const double x,    const GlFixed y)    { GlFixed t;  t.v = GlFixedMult(GlDoubleToFixed(x), y.v);        return t; }
	inline friend GlFixed operator *  (const GlFixed x, const double y)		 { GlFixed t;  t.v = GlFixedMult(x.v, GlDoubleToFixed(y));		return t; }

	inline friend GlFixed operator /  (const GlFixed x, const GlFixed y)    { GlFixed t;  t.v = GlFixedDiv(x.v, y.v);           return t; }
	inline friend GlFixed operator /  (const int       x, const GlFixed y)    { GlFixed t;  t.v = GlFixedDiv(GlIntToFixed(x), y.v);           return t; }
	inline friend GlFixed operator /  (const GlFixed x, const int       y)    { GlFixed t;  t.v = GlFixedDiv(x.v, GlIntToFixed(y));           return t; }

	inline friend GlFixed operator << (const GlFixed x, const int y)    { GlFixed t;  t.v = x.v << y;    return t; }
	inline friend GlFixed operator >> (const GlFixed x, const int y)    { GlFixed t;  t.v = x.v >> y;    return t; }

	inline friend bool operator == (const GlFixed x, const GlFixed y)    { return (x.v == y.v);       }
	inline friend bool operator == (const int x,       const GlFixed y)    { return (GlIntToFixed(x) == y.v);       }
	inline friend bool operator == (const GlFixed x, const int y)			{ return (x.v == GlIntToFixed(y));       }

	inline friend bool operator != (const GlFixed x, const GlFixed y)    { return (x.v != y.v);       }
	inline friend bool operator != (const int x,		  const GlFixed y)    { return (GlIntToFixed(x) != y.v);       }
	inline friend bool operator != (const GlFixed x, const int y)			{ return (x.v !=GlIntToFixed(y));       }

	inline friend bool operator <  (const GlFixed x, const GlFixed y)   { return (x.v < y.v);        }
	inline friend bool operator <  (const int x,	 const GlFixed y)   { return (GlIntToFixed(x) < y.v);        }
	inline friend bool operator <  (const GlFixed x, const int y)		{ return (x.v < GlIntToFixed(y) );        }

	inline friend bool operator >  (const GlFixed x, const GlFixed y)   { return (x.v > y.v);        }
	inline friend bool operator >  (const int x,	 const GlFixed y)   { return (GlIntToFixed(x) > y.v);        }
	inline friend bool operator >  (const GlFixed x, const int y)		{ return (x.v > GlIntToFixed(y));        }

	inline friend bool operator <= (const GlFixed x, const GlFixed y)    { return (x.v <= y.v);       }
	inline friend bool operator <= (const int x,		  const GlFixed y)    { return (GlIntToFixed(x) <= y.v);       }
	inline friend bool operator <= (const GlFixed x, const int y)			{ return (x.v <= GlIntToFixed(y));       }

	inline friend bool operator >= (const GlFixed x, const GlFixed y)    { return (x.v >= y.v);       }
	inline friend bool operator >= (const int x,		  const GlFixed y)    { return (GlIntToFixed(x) >= y.v);       }
	inline friend bool operator >= (const GlFixed x, const int y)			{ return (x.v >= GlIntToFixed(y));       }

    S32 v;
};


#endif
