/*--License:
	Kyra Sprite Engine
	Copyright Lee Thomason (Grinning Lizard Software) 2001-2002
	www.grinninglizard.com/kyra
	www.sourceforge.net/projects/kyra

	Kyra is provided under 2 licenses:

	- The GPL, with no additional restrictions.
	- The LGPL, provided you display the Kyra splash screen, described below.


--- GPL License --
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	The full text of the license can be found in license.txt


--- LGPL License --
  **Provided you kindly display the Kyra splash screen (details below), 
	you	may use the LGPL license:**

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	The full text of the license can be found in lgpl.txt


--- Kyra Splash Screen.

	It would be appreciate if you display the Kyra splash screen when using
	either license, however it is only required for the LGPL. All the
	resources for the splash are compiled into the library, and it can be
	accessed through the following API:

		KrEngine::StartSplash
		KrEngine::UpdateSplash
		KrEngine::EndSplash

	Full documentation is provided with the KrEngine class. The splash screen
	should be displayed for 2 seconds.

	Thank you.
*/

#ifndef GLMAP_INCLUDED
#define GLMAP_INCLUDED

#pragma warning( disable : 4530 )
#pragma warning( disable : 4786 )
#include <string>
#include <ctype.h>
#include "glutil.h"
#include "gldebug.h"
#include "glprime.h"
#include "gltypes.h"

// Standard Hash classes.
class GlHash
{
  public:
	U32 HashValue()	{ return val; }
  protected:
	U32 val;	
};

class GlStringHash : public GlHash
{
  public:
	GlStringHash( const std::string& str )
	{	
		val = 0;
		for ( unsigned i=0; i<str.length() && i < 32; ++i )
		{
			val <<= 1;
			val |= str[i];
		}
	}
};

template < class T >
class GlNumberHash : public GlHash
{
  public:
	GlNumberHash( T num )	{ val = (T) num; }
};


// A map template class. It associates a KEY with a VALUE.
// It uses a HASH class to get a hash value from the KEY. 2 Hash classes are
// provided: GlStringHash and GlNumberHash.

template < class KEY, class VALUE, class HASH >
class GlMap
{
  public:
	GlMap( U32 startSize = 64, U32 grow = 60 );
	~GlMap();

	// Adds a key value pair. Will fail if the key is not unique.
	bool Add( const KEY& key, const VALUE& value );
	
	// Remove a key value pair. Will fail if the key is not found.
	bool Remove( const KEY& key );
	
	// Returns true if the key is found, and writes the result to value.
	// Value will not be set if the key is not found.
	bool Find( const KEY& key, VALUE* value );

	// PRIVATE! But silly template friends too hard.
	enum {
		EXPAND = 4
	};

	struct Item
	{
		KEY		key;
		VALUE	value;
		Item*	next;
	};

	// Grow the hash table. Dicey work, and slow.
	void Grow( unsigned newsize );

	U32 numBuckets;
	Item** buckets;
	U32 grow;
	U32 numItems;
};




template < class KEY, class VALUE, class HASH >
inline GlMap< KEY, VALUE, HASH>::GlMap( U32 startSize, U32 _grow )
{
	numBuckets = GlPrime( startSize, 1 );
	buckets = new Item*[ numBuckets ];
	memset( buckets, 0, sizeof( Item* ) * numBuckets );
	grow = _grow;
	numItems = 0;

	#ifdef DEBUG
		//GLOUTPUT( "Created Map: %d buckets\n", numBuckets );
	#endif
};


template < class KEY, class VALUE, class HASH >
inline GlMap< KEY, VALUE, HASH>::~GlMap()
{
	for( U32 i=0; i<numBuckets; ++i )
	{	
		while ( buckets[i] )
		{
			Item* next = buckets[i]->next;
			delete buckets[i];
			buckets[i] = next;
		}
	}
		
	// Everything is in the new list, create new buckets and put it back in.
	delete [] buckets;
};


template < class KEY, class VALUE, class HASH >
inline void GlMap< KEY, VALUE, HASH>::Grow( unsigned newsize )
{
	#ifdef DEBUG
	int itemcount = 0;
	#endif

	// Yep. Unlink everything, put in a linked list, and re-insert.
	Item *root = 0;
	for( U32 i=0; i<numBuckets; ++i )
	{	
		while ( buckets[i] )
		{
			Item* next = buckets[i]->next;

			buckets[i]->next = root;
			root = buckets[i];

			buckets[i] = next;

			#ifdef DEBUG
			++itemcount;
			#endif
		}
	}
	#ifdef DEBUG
	int comparecount = 0;
	for( Item* it = root; it; it = it->next )
		++comparecount;
	GLASSERT( comparecount == itemcount );
	#endif
		
	// Everything is in the new list, create new buckets and put it back in.
	delete [] buckets;
	buckets = 0;

	#ifdef DEBUG
		//GLOUTPUT( "Rebuilding map (from %d)...", numBuckets );
	#endif

	numBuckets = GlPrime( newsize, 1 );
	#ifdef DEBUG
		//GLOUTPUT( "%d buckets\n", numBuckets );
	#endif

	buckets = new Item*[ numBuckets ];
	memset( buckets, 0, sizeof( Item* ) * numBuckets );

	while ( root )
	{
		Item* next = root->next;

		HASH hash( root->key );
		U32 which = hash.HashValue() % numBuckets;

		root->next = buckets[ which ];
		buckets[ which ] = root;

		root = next;
	}
}


template < class KEY, class VALUE, class HASH >
inline bool GlMap< KEY, VALUE, HASH>::Add( const KEY& key, const VALUE& value )
{
	VALUE dummy;
	if ( Find( key, &dummy ) ) 
		return false;

	// Do we need to get bigger?
	if ( ( ( numItems + 1 ) * 100 / numBuckets ) > grow )
	{
		Grow( GlMax( numItems * EXPAND, numBuckets * EXPAND ) );
	}

	// Add the key in, if it is unique.
	HASH hash( key );
	U32 which = hash.HashValue() % numBuckets;

	Item* item	= new Item;
	item->key	= key;
	item->value = value;
	item->next	= buckets[ which ];

	buckets[ which ] = item;
	++numItems;
	return true;
}


template < class KEY, class VALUE, class HASH >
inline bool GlMap< KEY, VALUE, HASH>::Find( const KEY& key, VALUE* value )
{
	HASH hash( key );
	U32 which = hash.HashValue() % numBuckets;

	Item* item = buckets[which];
	
	while( item )
	{
		if ( item->key == key )
		{
			*value = item->value;
			return true;
		}
		item = item->next;
	}
	return false;	
}


template < class KEY, class VALUE, class HASH >
inline bool GlMap< KEY, VALUE, HASH>::Remove( const KEY& key )
{
	HASH hash( key );
	U32 which = hash.HashValue() % numBuckets;

	Item* item = buckets[which];
	Item* prev = 0;
	
	while( item )
	{
		if ( item->key == key )
		{
			if ( prev )
				prev->next = item->next;
			else
				buckets[ which ] = item->next;

			delete item;
			--numItems;
			return true;
		}
		prev = item;
		item = item->next;
	}
	return false;	
}


template < class KEY, class VALUE, class HASH >
class GlMapIterator
{
  public:
	GlMapIterator( GlMap< KEY, VALUE, HASH >& _map )	{ map = &_map; bucket=0; item=0; }

	void Begin()		{ bucket=0; item=map->buckets[0]; FindValid(); }
	void Next()			{ item=item->next; FindValid(); }
	bool Done()			{ return ( bucket < 0 ); }

	void Current( KEY* key, VALUE* value )
	{
		if ( item )
		{
			*key = item->key;
			*value = item->value;
		}
	}

  private:
	GlMap< KEY, VALUE, HASH >* map;
	int bucket;
	typename GlMap< KEY, VALUE, HASH >::Item* item;

	void FindValid()
	{
		if (	bucket < 0				// we are done
			 || ( item && bucket >=0 ))	// we have a current value
		{
			return;
		}
		GLASSERT( item == 0 );			// else we should have current value
		for ( ++bucket; bucket < map->numBuckets; ++bucket )
		{
			if ( map->buckets[bucket] )
			{
				item = map->buckets[ bucket ];
				return;
			}
		}
		bucket = -1;
		item = 0;
	}
};


#endif

